/*****************************************************************************
 ***      Author:  Robbie Gubler -- tarsin@happybox.org                    ***
 ***        Date:  2001.04.28                                              ***
 *** Description:  TEA(Tiny Encryption Algorithm) Implementation.          ***
 ***    Comments:  The TEA is a Feistel type cipher; the block is split    ***
 ***               into two parts, each being 32 bits, 128 bit key, and by ***
 ***               default uses 32 rounds.  The cipher was orginally       ***
 ***               written by David Wheeler and later modifed by Roger M.  ***
 ***               Needham.                                                ***
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TEA.h"
#include "TEA-ICommandLine.h"

#define TRUE             1
#define FALSE           -1
#define ROUNDS          32
#define DECIPHER        'd'
#define ENCIPHER        'e'

static void __strrev(char* l) {
    char* r = l + strlen(l) - 1;
    char temp = '\0';
    while(l < r) {
        temp = *l;
        *l = *r;
        *r = temp;
        *l++;
        *r--;
    }
}
static char* ltoa(long val, char *buf, int radix) {
	xtoa((unsigned long)val, buf, radix, (radix == 10 && val < 0));
        return buf;
}

static void xtoa(unsigned long val, char *buf, unsigned radix, int is_neg) {
        char *p;                /* pointer to traverse string */
        char *firstdig;         /* pointer to first digit */
        char temp;              /* temp char */
        unsigned digval;        /* value of digit */

        p = buf;

        if (is_neg) {
            /* negative, so output '-' and negate */
            *p++ = '-';
            val = (unsigned long)(-(long)val);
        }

        firstdig = p;           /* save pointer to first digit */

        do {
            digval = (unsigned) (val % radix);
            val /= radix;       /* get next digit */

            /* convert to ascii and store */
            if (digval > 9)
                *p++ = (char) (digval - 10 + 'a');  /* a letter */
            else
                *p++ = (char) (digval + '0');       /* a digit */
        } while (val > 0);

        /* We now have the digit of the number in the buffer, but in reverse
           order.  Thus we reverse them now. */

        *p-- = '\0';            /* terminate string; p points to last digit */

        do {
            temp = *p;
            *p = *firstdig;
            *firstdig = temp;   /* swap *p and *firstdig */
            --p;
            ++firstdig;         /* advance to next two digits */
        } while (firstdig < p); /* repeat until halfway */
}

int ReadFile(int type, unsigned long* key, char* fi_name) {
    signed long fi_size;
    int i = 0, x = 0;
    unsigned long data[2];
    char* temp;
    char* str_int;
    char* fo_name = malloc(strlen(fi_name) + 10);
    FILE* fi;
    FILE* fo;

    fi = fopen(fi_name, "rb");
    if(fi == NULL) {
        temp = calloc(1, (strlen(fi_name)+64));
        strcpy(temp, "Error: Can't read from file \"");
        strncat(temp, fi_name, strlen(fi_name));
        strcat(temp, "\"");
        StdOutDisplay(temp);
        return TRUE;
    } else {
        if(type == ENCIPHER) {
            strncpy(fo_name, fi_name, strlen(fi_name)+1);
            strncat(fo_name, ".tea-enc", 8);
        } else if(type == DECIPHER) {
            if(strlen(fi_name) > 8) {
                temp = malloc(strlen(fi_name)+1);
                __strrev(fi_name);
                strncpy(temp, fi_name, 8);
                temp[8] = '\0';
                __strrev(fi_name);
                __strrev(temp);
                if(strcmp(temp, ".tea-enc") == 0) {
                    strncpy(fo_name, fi_name, (strlen(fi_name)-8));
                    fo_name[strlen(fi_name)-8] = '\0';
                    strncat(fo_name, ".tea-dec\0", 9);
                }
            }    
        }
        fo = fopen(fo_name, "wb");
        if(fo == NULL) {
            temp = calloc(1, (strlen(fo_name)+64));
            strcpy(temp, "Error: Can't write to file \"");
            strncat(temp, fo_name, strlen(fo_name));
            strcat(temp, "\"");
            StdOutDisplay(temp);
            return TRUE; 
        }
    }
    fseek(fi, 1, SEEK_END);
    x = fi_size = (ftell(fi))-1;
    rewind(fi);

    if(type == ENCIPHER) {
        temp = calloc(1, (strlen(fi_name)+64));
        str_int = calloc(1, 16);
        strcpy(temp, "Encrypting a total of ");
        strcat(temp, ltoa(fi_size, str_int, 10));
        strcat(temp, " bytes from \"");
        strncat(temp, fi_name, strlen(fi_name));
        strcat(temp, "\"");
        StdOutDisplay(temp);
    } else {
        temp = calloc(1, (strlen(fi_name)+64));
        str_int = calloc(1, 32);
        strcpy(temp, "Decrypting a total of ");
        strcat(temp, ltoa(fi_size, str_int, 10));
        strcat(temp, " bytes from \"");
        strncat(temp, fi_name, strlen(fi_name));
        strcat(temp, "\"");
        StdOutDisplay(temp);
    }
    for(i = 0; fi_size >= 1; (fi_size -= 8)) {
        char* tmp;
        int n;
		int discard;

        memset(data, 0, (sizeof(unsigned long)*2));
        fread(data, 4, 2, fi);
        if(type == ENCIPHER) {
            Encipher(data, key, ROUNDS);
            fwrite(data, 4, 2, fo);
        } else if(type == DECIPHER) {
            Decipher(data, key, ROUNDS);
            if(fi_size <= 8) {
                tmp = (char*)&data;
                for(n = 0, discard = FALSE; n < 8 && discard == FALSE; n++) {
                    if(tmp[n] == '\0') {
						discard = TRUE;
                    }
                }
				fwrite(data, n-1, 1, fo);
            } else                   
				fwrite(data, 4, 2, fo);

			/*
            Decipher(data, key, ROUNDS);
            fwrite(data, 4, 2, fo);
*/

        }
    }
    
    if(type == ENCIPHER) {
        temp = calloc(1, (strlen(fo_name)+64));
        strcpy(temp, "Ciphertext written to \"");
        strncat(temp, fo_name, strlen(fo_name));
        strcat(temp, "\"");
        StdOutDisplay(temp);
    } else {
        temp = calloc(1, (strlen(fo_name)+64));
        strcpy(temp, "Ciphertext written to \"");
        strncat(temp, fo_name, strlen(fo_name));
        strcat(temp, "\"");
        StdOutDisplay(temp);

    }

    return FALSE;
}

int GetPassPhraseFromFile(unsigned long* key, char* filename, unsigned long start_line, unsigned long phrase_size) {
    unsigned char* phrase = malloc(phrase_size+1);
    unsigned long n;
    char ch;
    FILE* fi;

    fi = fopen(filename, "rb");
    if(fi == NULL) {
        fprintf(stderr, "\nError:  Can't open file(%s)", filename);
        return TRUE;
    }
    
    for(n = 1; n < start_line;) {
        ch = fgetc(fi);
        n = (ch == '\n') ? n+1: n;
    }

    fread(phrase, phrase_size, 1, fi);
    fclose(fi);
    phrase[phrase_size] = '\0';
    
    KeyCrunch(key, phrase);

    return FALSE;

}

int KeyCrunch(unsigned long* key, const unsigned char* str) {
    unsigned long phrase_size;
    unsigned long one_fourth;
    unsigned long n;

    if(strlen(str) < 8)
        return TRUE;

    phrase_size = strlen(str);
    one_fourth = phrase_size / 4;

    key[0] = HashAlgorithm(str, one_fourth);
    for(n = 0; n <= one_fourth; ++n, ++str);
    key[1] = HashAlgorithm(str, one_fourth);
    for(n = 0; n <= one_fourth; ++n, ++str);
    key[2] = HashAlgorithm(str, one_fourth);
    for(n = 0; n <= ((phrase_size)-(one_fourth*3)); ++n, ++str);
    key[3] = HashAlgorithm(str, ((phrase_size)-(one_fourth*3)));

    return FALSE;

}

unsigned long HashAlgorithm(const unsigned char* str, int size) {
    unsigned long h = 0, g;
    
    while(*str) {
        h = (h << 4) + *str++;
        if(g = h & 0xF0000000)
            h ^= g >> 24;

        h &= ~g;
    }

    return h;
}

void Encipher(unsigned long* v, unsigned long* k, unsigned long n) {
    unsigned long y = v[0],
                  z = v[1],
                  sum = 0,
                  delta = 0x9e3779b9,
                  limit = delta*n;

    while(sum != limit) {
        y   += ((z<<4) ^ (z>>5)) + (z^sum) + k[sum&3];
        sum += delta;
        z   += ((y<<4) ^ (y>>5)) + (y^sum) + k[(sum>>11)&3];
    }

    v[0] = y;
    v[1] = z;
}

void Decipher(unsigned long* v, unsigned long* k, unsigned long n) {
    unsigned long y = v[0],
                  z = v[1],
                  delta = 0x9e3779b9,
                  sum = delta*(-n);
    sum = delta << 5;
    while(sum) {
        z   -= ((y<<4) ^ (y>>5)) + (y^sum) + k[(sum>>11)&3];
        sum -= delta;
        y   -= ((z<<4) ^ (z>>5)) + (z^sum) + k[sum&3];
    }

    v[0] = y;
    v[1] = z;
}

